//-----------------------------------------------------------------------------
// File: textures.cpp
//
// Desc: display a perlin texture 
//
// author : adrien.auclair
//	        adrien.auclair@club-internet.fr
// date   : 21/09/2004
//-----------------------------------------------------------------------------
#include <Windows.h>
#include <mmsystem.h>
#include <d3dx9.h>
#include <dxerr9.h>
#include <assert.h>

#include "../my_common/util.h"
#include "perlin_texture.h"

#define SAFE_RELEASE(x) { if (x!=NULL) { x->Release(); x=NULL; } }

//-----------------------------------------------------------------------------
// Global variables
//-----------------------------------------------------------------------------
LPDIRECT3D9             g_pD3D                 = NULL; // Used to create the D3DDevice
LPDIRECT3DDEVICE9       g_pd3dDevice           = NULL; // Our rendering device
LPDIRECT3DVERTEXBUFFER9 g_pVB_Floor            = NULL; // Buffer to hold vertices for the wall

//LPDIRECT3DTEXTURE9      g_pTexture             = NULL; // The wall texture
LPDIRECT3DTEXTURE9      g_pPerlinTexture        = NULL; // The floor texture

LPDIRECT3DVERTEXSHADER9      m_pVertexShader      = NULL;  // vertex shader
LPDIRECT3DVERTEXDECLARATION9 m_pVertexDeclaration = NULL;  // vertex shader input data type
LPD3DXCONSTANTTABLE          m_pVertexConstants   = NULL;  // array of vertex shader constants

LPDIRECT3DPIXELSHADER9       m_pPixelShader       = NULL;  // pixel shader
LPD3DXCONSTANTTABLE          m_pPixelConstants    = NULL;  // array of pixel shader constant


D3DXMATRIXA16 matView;
D3DXMATRIXA16 matWorld;    
D3DXMATRIXA16 matProj;    
D3DXVECTOR3   vEyePt( 0.5f, 1.4f,0.5f );
D3DXVECTOR3 vLookatPt( 0.5f, 0.5f, 0.5f );
D3DXVECTOR3 vUpVec( 0.0f, 0.0f, 1.0f );    
D3DXMATRIX  MFull,TMFull; // Full = World * View * Projection


// A structure for our custom vertex type
struct CUSTOMVERTEX
{
    FLOAT x, y, z;      // The untransformed, 3D position for the vertex		
	FLOAT tu, tv;       // The texture coordinates	
};


//-----------------------------------------------------------------------------
// Name: InitShaders()
// Desc: Initialize vertex and pixel shaders
//-----------------------------------------------------------------------------
HRESULT InitShaders()
{
	//--------------------------- creation du vertex shader
	DWORD dwFlags = 0; 
    dwFlags |= D3DXSHADER_DEBUG;
    LPD3DXBUFFER pCode   = NULL;
    LPD3DXBUFFER pErrMsg = NULL;
	HRESULT       s;

	// array describing the data type :
	D3DVERTEXELEMENT9 decl[] =
    {
      { 0,  0, D3DDECLTYPE_FLOAT3, D3DDECLMETHOD_DEFAULT, D3DDECLUSAGE_POSITION, 0 },      	  
	  { 0, 12, D3DDECLTYPE_FLOAT2, D3DDECLMETHOD_DEFAULT, D3DDECLUSAGE_TEXCOORD, 0 },	  	  
        D3DDECL_END()
    };

	// the type of data passed to the vertex shader is created
    s = g_pd3dDevice->CreateVertexDeclaration(decl,&m_pVertexDeclaration);
    if(FAILED(s))
    {
		MessageBox(NULL,"CreateVertexDeclaration failed","CRendererDX9::Create",MB_OK|MB_ICONEXCLAMATION);
		return false;
    }

	// compile the HLSL source
	s=D3DXCompileShaderFromFile("shaders/vs_pass.cg",NULL,NULL,"main","vs_2_0",dwFlags,&pCode,&pErrMsg,&m_pVertexConstants);
	if( FAILED(s))
	{
		if (pErrMsg != NULL) MessageBox(NULL, (char *)(pErrMsg->GetBufferPointer()), "HLSL compilation error", MB_OK);
			else MessageBox(NULL,DXGetErrorString9(s), "HLSL compilation error", MB_OK);
        return E_FAIL;
	}

	// create the shader from the compiled code
	char* szCode = (char*)pCode->GetBufferPointer();
    s = g_pd3dDevice->CreateVertexShader((DWORD*)pCode->GetBufferPointer(),&m_pVertexShader);
    pCode->Release();
    if(FAILED(s))
    {
		MessageBox(NULL,"CreateVertexShader failed","CRendererDX9::Create",MB_OK|MB_ICONEXCLAMATION);
		return E_FAIL;		
    }

	//------------------ creation du pixel shader
	// compile the source code
	pErrMsg = NULL;
    s = D3DXCompileShaderFromFile("shaders/ps_pass.cg",NULL,NULL,"main","ps_2_0",dwFlags,&pCode,&pErrMsg,&m_pPixelConstants);
	if( FAILED(s))
	{
		if (pErrMsg != NULL) MessageBox(NULL, (char *)(pErrMsg->GetBufferPointer()), "HLSL compilation error", MB_OK);
			else MessageBox(NULL,DXGetErrorString9(s), "HLSL compilation error", MB_OK);
        return E_FAIL;
	}    

	// create the pixel shader from the copiled code    
    s  = g_pd3dDevice->CreatePixelShader((DWORD*)pCode->GetBufferPointer(),&m_pPixelShader);
    pCode->Release();
    if(FAILED(s))
    {
		MessageBox(NULL,"CreatePixelShader failed","CRendererDX9::Create",MB_OK|MB_ICONEXCLAMATION);
		return false;
    }	


	return S_OK;
}


//-----------------------------------------------------------------------------
// Name: InitD3D()
// Desc: Initializes Direct3D
//-----------------------------------------------------------------------------
HRESULT InitD3D( HWND hWnd )
{
    // Create the D3D object.
    if( NULL == ( g_pD3D = Direct3DCreate9( D3D_SDK_VERSION ) ) )
        return E_FAIL;

    // Set up the structure used to create the D3DDevice
    D3DPRESENT_PARAMETERS d3dpp;
    ZeroMemory( &d3dpp, sizeof(d3dpp) );
    d3dpp.Windowed = TRUE;
    d3dpp.SwapEffect = D3DSWAPEFFECT_DISCARD;
    d3dpp.BackBufferFormat = D3DFMT_UNKNOWN;
	d3dpp.BackBufferCount = 1;
	d3dpp.hDeviceWindow = hWnd;
	d3dpp.EnableAutoDepthStencil = TRUE;
    d3dpp.AutoDepthStencilFormat = D3DFMT_D16;
	d3dpp.PresentationInterval = D3DPRESENT_INTERVAL_IMMEDIATE;

    // Create the D3DDevice
    if( FAILED( g_pD3D->CreateDevice( D3DADAPTER_DEFAULT, D3DDEVTYPE_HAL, hWnd,
                                      D3DCREATE_SOFTWARE_VERTEXPROCESSING,
                                      &d3dpp, &g_pd3dDevice ) ) )
    {
		MessageBox(NULL, "Could not create the device", "Textures.exe", MB_OK);
        return E_FAIL;
    }

    // Turn off culling, so we see the front and back of the triangle
    g_pd3dDevice->SetRenderState( D3DRS_CULLMODE, D3DCULL_NONE );

    // Turn off D3D lighting, since we are providing our own vertex colors
    g_pd3dDevice->SetRenderState( D3DRS_LIGHTING, FALSE );	
	
	
	if( FAILED(InitShaders() ))
		return E_FAIL;

	
    if FAILED(InitFont()) 
		return E_FAIL;

    return S_OK;
}



//-----------------------------------------------------------------------------
// Name: InitGeometry()
// Desc: Creates the scene geometry
//-----------------------------------------------------------------------------
HRESULT InitGeometry()
{
	g_pPerlinTexture = createPerlinTexture(g_pd3dDevice , 256 , 256);	


	//----------------- FLOOR	
	// format is      x,y,z     texture 
    CUSTOMVERTEX g_Vertices2[] =
    {
        {  0.0f, 0.0f, 0.0f			, 0.0f , 1.0f  },  //1
        {  1.0f, 0.0f, 1.0f			, 1.0f , 0.0f  },   //2
        {  0.0f, 0.0f, 1.0f			, 0.0f , 0.0f  },   //3

		{  0.0f, 0.0f, 0.0f			, 0.0f , 1.0f  },  //4
        {  1.0f, 0.0f, 0.0f			, 1.0f , 1.0f  },   //5
        {  1.0f, 0.0f, 1.0f			, 1.0f , 0.0f  }    //6		
    };

    // Create the vertex buffer.
    if( FAILED( g_pd3dDevice->CreateVertexBuffer( 6*sizeof(CUSTOMVERTEX),
                                                  0, 0, // on se passe de D3DFVF_CUSTOMVERTEX,
                                                  D3DPOOL_MANAGED, &g_pVB_Floor, NULL ) ) )
    {
		MessageBox(NULL, "Could not create vertex buffer", "Textures.exe", MB_OK);
        return E_FAIL;
    }

	VOID* pVertices;
    // Fill the vertex buffer.    
    if( FAILED( g_pVB_Floor->Lock( 0, sizeof(g_Vertices2), (void**)&pVertices, 0 ) ) )
        return E_FAIL;
    memcpy( pVertices, g_Vertices2, sizeof(g_Vertices2) );
    g_pVB_Floor->Unlock();

	
    return S_OK;
}



//-----------------------------------------------------------------------------
// Name: Cleanup()
// Desc: Releases all previously initialized objects
//-----------------------------------------------------------------------------
VOID Cleanup()
{		

	SAFE_RELEASE(m_pVertexShader);		
	SAFE_RELEASE(m_pVertexDeclaration);
	SAFE_RELEASE(m_pVertexConstants);	
	SAFE_RELEASE(m_pPixelShader);
	SAFE_RELEASE(m_pPixelConstants);		
	SAFE_RELEASE(g_pVB_Floor);		
	SAFE_RELEASE(g_pd3dDevice);
	SAFE_RELEASE(g_pD3D);	
}





//-----------------------------------------------------------------------------
// Name: SetupMatrices()
// Desc: Sets up the world, view, and projection transform Matrices.
//-----------------------------------------------------------------------------
VOID SetupMatrices()
{     
    // Set up our view matrix.      
    D3DXMatrixLookAtLH( &matView, &vEyePt, &vLookatPt, &vUpVec );    

    // set up our projection matrix
    D3DXMATRIXA16 matProj;
    D3DXMatrixPerspectiveFovLH( &matProj, D3DX_PI/4, 1.0f, 1.0f, 100.0f );    
	
	MFull = matWorld * matView * matProj;
	D3DXMatrixTranspose(&TMFull,&MFull);
}

//-----------------------------------------------------------------------------
// Name: DrawFloor()
// Desc: Draws the floor (texture)
//-----------------------------------------------------------------------------
void Draw_Floor()
{
		D3DXMatrixIdentity(&matWorld);
		// Setup the world, view, and projection Matrices
        SetupMatrices();	

		g_pd3dDevice->SetVertexDeclaration(m_pVertexDeclaration);
		g_pd3dDevice->SetVertexShader     (m_pVertexShader);
		g_pd3dDevice->SetPixelShader      (m_pPixelShader);

	    // set the constant parameter modelViewProj matrix for the vertrex shader
		D3DXHANDLE handle;
        if(handle = m_pVertexConstants->GetConstantByName(NULL,"modelViewProj"))
			m_pVertexConstants->SetMatrix(g_pd3dDevice,handle,(D3DXMATRIX*)&TMFull);

		int TextureIndex=0;

		if(m_pPixelConstants && (handle = m_pPixelConstants->GetConstantByName(NULL,"decal")))
		{
			D3DXCONSTANT_DESC constDesc;
			UINT count = 1;
			m_pPixelConstants->GetConstantDesc(handle,&constDesc,&count);

			if(constDesc.RegisterSet == D3DXRS_SAMPLER)
			{
				TextureIndex = constDesc.RegisterIndex;
					
			}
			g_pd3dDevice->SetTexture( TextureIndex, g_pPerlinTexture );
			
		}
		
		
		// render the texture
        g_pd3dDevice->SetStreamSource( 0, g_pVB_Floor, 0, sizeof(CUSTOMVERTEX) );        		
        g_pd3dDevice->DrawPrimitive( D3DPT_TRIANGLELIST, 0, 2 );		
		
}

//-----------------------------------------------------------------------------
// Name: Render()
// Desc: Draws the scene
//-----------------------------------------------------------------------------
VOID Render()
{	
    // Clear the backbuffer to a black color
    g_pd3dDevice->Clear( 0, NULL, D3DCLEAR_TARGET| D3DCLEAR_ZBUFFER, D3DCOLOR_XRGB(0,0,0), 1.0f, 0 );	
	

    // Begin the scene
    if( SUCCEEDED( g_pd3dDevice->BeginScene() ) )
    {	
				
		Draw_Floor();
		drawFPS(0,white);
        // End the scene
        g_pd3dDevice->EndScene();
    }

    // Present the backbuffer contents to the display
    g_pd3dDevice->Present( NULL, NULL, NULL, NULL );
}




//-----------------------------------------------------------------------------
// Name: MsgProc()
// Desc: The window's message handler
//-----------------------------------------------------------------------------
LRESULT WINAPI MsgProc( HWND hWnd, UINT msg, WPARAM wParam, LPARAM lParam )
{
    switch( msg )
    {
        case WM_DESTROY:
            Cleanup();
            PostQuitMessage( 0 );
            return 0;
    }

    return DefWindowProc( hWnd, msg, wParam, lParam );
}




//-----------------------------------------------------------------------------
// Name: WinMain()
// Desc: The application's entry point
//-----------------------------------------------------------------------------
INT WINAPI WinMain( HINSTANCE hInst, HINSTANCE, LPSTR, INT )
{
    // Register the window class
    WNDCLASSEX wc = { sizeof(WNDCLASSEX), CS_CLASSDC, MsgProc, 0L, 0L,
                      GetModuleHandle(NULL), NULL, NULL, NULL, NULL,
                      "HLSL basics", NULL };
    RegisterClassEx( &wc );

    // Create the application's window
    HWND hWnd = CreateWindow( "HLSL basics", "HLSL basics : texture",
                              WS_OVERLAPPEDWINDOW, 100, 100, 800, 600,
                              GetDesktopWindow(), NULL, wc.hInstance, NULL );
	
    // Initialize Direct3D
    if( SUCCEEDED( InitD3D( hWnd ) ) )
    {
        // Create the scene geometry
        if( SUCCEEDED( InitGeometry() ) )
        {	
			InitTimer();
            // Show the window
            ShowWindow( hWnd, SW_SHOWDEFAULT );
            UpdateWindow( hWnd );
            // Enter the message loop
            MSG msg;
            ZeroMemory( &msg, sizeof(msg) );
            while( msg.message!=WM_QUIT )
            {
                if( PeekMessage( &msg, NULL, 0U, 0U, PM_REMOVE ) )
                {
                    TranslateMessage( &msg );
                    DispatchMessage( &msg );
                }
                else
				{
					ctime = timeGetTime();
                    Render();					
				}
            }
        }
    }

    UnregisterClass( "D3D Tutorial", wc.hInstance );
    return 0;
}



